@@ -18,6 +18,8 @@
 
 #include "goxel.h"
 
+#include "../ext_src/lua/lauxlib.h"
+
 typedef struct {
     UT_hash_handle  hh;
     action_t        action;
@@ -39,11 +41,14 @@ void action_register(const action_t *action)
     HASH_ADD_KEYPTR(hh, g_actions, action->id, strlen(action->id), item);
 }
 
-action_t *action_get(const char *id)
+action_t *action_get(const char *id, bool assert_exists)
 {
     action_hash_item_t *item;
     HASH_FIND_STR(g_actions, id, item);
-    if (!item) LOG_W("Cannot find action %s", id);
+    if (!item && assert_exists) {
+        LOG_E("Cannot find action %s", id);
+        assert(false);
+    }
     return item ? &item->action : NULL;
 }
 
@@ -55,51 +60,80 @@ void actions_iter(int (*f)(action_t *action, void *user), void *user)
     }
 }
 
+static const void *topointer(lua_State *l, int idx)
+{
+    if (lua_isstring(l, idx))
+        return lua_tostring(l, idx);
+    if (lua_islightuserdata(l, idx))
+        return lua_topointer(l, idx);
+    return *(void**)lua_touserdata(l, idx);
+}
+
 // The default action function will try to call the action cfunc.
-static int default_function(const action_t *a, astack_t *s)
+static int default_function(const action_t *a, lua_State *l)
 {
     int i;
-    void (*func)() = a->cfunc;
+    void *p;
     assert(a->cfunc);
     assert(a->csig);
 
     // Fill any missing arguments with 0 values.
-    for (i = stack_size(s); i < strlen(a->csig + 1); i++) {
+    for (i = lua_gettop(l); i < strlen(a->csig + 1); i++) {
         switch (a->csig[i + 1]) {
-            case 'i': stack_push_i(s, 0); break;
-            case 'p': stack_push_p(s, NULL); break;
+            case 'i': lua_pushnumber(l, 0); break;
+            case 'p': lua_pushlightuserdata(l, NULL); break;
             default: assert(false);
         }
     }
 
+    // XXX: clean this up.
     if (strcmp(a->csig, "v") == 0) {
+        void (*func)(void) = a->cfunc;
         func();
     } else if (strcmp(a->csig, "vp") == 0) {
-        func(stack_get_p(s, 0));
+        void (*func)(const void *) = a->cfunc;
+        func(topointer(l, 1));
     } else if (strcmp(a->csig, "vpp") == 0) {
-        func(stack_get_p(s, 0),
-             stack_get_p(s, 1));
+        void (*func)(const void *, const void *) = a->cfunc;
+        func(topointer(l, 1),
+             topointer(l, 2));
     } else if (strcmp(a->csig, "vpi") == 0) {
-        func(stack_get_p(s, 0),
-             stack_get_i(s, 1));
+        void (*func)(const void *, int) = a->cfunc;
+        func(topointer(l, 1),
+             lua_tointeger(l, 2));
     } else if (strcmp(a->csig, "vppp") == 0) {
-        func(stack_get_p(s, 0),
-             stack_get_p(s, 1),
-             stack_get_p(s, 2));
+        void (*func)(const void *, const void *, const void *) = a->cfunc;
+        func(topointer(l, 1),
+             topointer(l, 2),
+             topointer(l, 3));
     } else if (strcmp(a->csig, "vppi") == 0) {
-        func(stack_get_p(s, 0),
-             stack_get_p(s, 1),
-             stack_get_i(s, 2));
+        void (*func)(const void *, const void *, int) = a->cfunc;
+        func(topointer(l, 1),
+             topointer(l, 2),
+             lua_tointeger(l, 3));
     } else if (strcmp(a->csig, "vpii") == 0) {
-        func(stack_get_p(s, 0),
-             stack_get_i(s, 1),
-             stack_get_i(s, 2));
+        void (*func)(const void *, int, int) = a->cfunc;
+        func(topointer(l, 1),
+             lua_tointeger(l, 2),
+             lua_tointeger(l, 3));
+    } else if (strcmp(a->csig, "p") == 0) {
+        void *(*func)(void) = a->cfunc;
+        p = func();
+        lua_pushlightuserdata(l, p);
+    } else if (strcmp(a->csig, "pp") == 0) {
+        void *(*func)(const void *) = a->cfunc;
+        p = func(topointer(l, 1));
+        lua_pushlightuserdata(l, p);
+    } else if (strcmp(a->csig, "ip") == 0) {
+        int (*func)(const void *) = a->cfunc;
+        i = func(topointer(l, 1));
+        lua_pushinteger(l, i);
     } else {
         LOG_E("Cannot handle sig '%s'", a->csig);
         assert(false);
     }
 
-    return 0;
+    return a->csig[0] == 'v' ? 0 : 1;
 }
 
 int action_execv(const action_t *action, const char *sig, va_list ap)
@@ -110,8 +144,8 @@ int action_execv(const action_t *action, const char *sig, va_list ap)
     char c;
     bool b;
     int i, nb;
-    int (*func)(const action_t *a, astack_t *s);
-    astack_t *s = stack_create();
+    int (*func)(const action_t *a, lua_State *l);
+    lua_State *l = luaL_newstate();
     func = action->func ?: default_function;
 
     if (reentry == 0 && (action->flags & ACTION_TOUCH_IMAGE)) {
@@ -123,38 +157,41 @@ int action_execv(const action_t *action, const char *sig, va_list ap)
     while ((c = *sig++)) {
         if (c == '>') break;
         switch (c) {
-            case 'i': stack_push_i(s, va_arg(ap, int)); break;
-            case 'b': stack_push_b(s, va_arg(ap, int)); break;
-            case 'p': stack_push_p(s, va_arg(ap, void*)); break;
+            case 'i': lua_pushnumber(l, va_arg(ap, int)); break;
+            case 'b': lua_pushboolean(l, va_arg(ap, int)); break;
+            case 'p': lua_pushlightuserdata(l, va_arg(ap, void*)); break;
             default: assert(false);
         }
     }
 
-    if ((action->flags & ACTION_TOGGLE) && (stack_size(s) == 0) && (!c)) {
-        func(action, s);
-        b = stack_get_b(s, 0);
-        stack_clear(s);
-        stack_push_b(s, !b);
+    if ((action->flags & ACTION_TOGGLE) && (lua_gettop(l) == 0) && (!c)) {
+        func(action, l);
+        b = lua_toboolean(l, 1);
+        lua_settop(l, 0);
+        lua_pushboolean(l, !b);
     }
-    func(action, s);
+    func(action, l);
 
     // Get the return arguments.
     nb = c ? (int)strlen(sig) : 0;
     for (i = 0; i < nb; i++) {
         c = sig[i];
         switch (c) {
-            case 'i': *va_arg(ap, int*) = stack_get_i(s, -nb + i); break;
-            case 'b': *va_arg(ap, bool*) = stack_get_b(s, -nb + i); break;
-            case 'p': *va_arg(ap, void**) = stack_get_p(s, -nb + i); break;
+            case 'i':
+                *va_arg(ap, int*) = lua_tointeger(l, -nb + i + 1);
+                break;
+            case 'b':
+                *va_arg(ap, bool*) = lua_toboolean(l, -nb + i + 1);
+                break;
+            case 'p':
+                *va_arg(ap, void**) = lua_touserdata(l, -nb + i + 1);
+                break;
             default: assert(false);
         }
     }
 
     reentry--;
-    if (reentry == 0 && (action->flags & ACTION_TOUCH_IMAGE)) {
-        goxel_update_meshes(-1);
-    }
-    stack_delete(s);
+    lua_close(l);
     return 0;
 }
 
@@ -167,3 +204,10 @@ int action_exec(const action_t *action, const char *sig, ...)
     va_end(ap);
     return ret;
 }
+
+int action_exec_lua(const action_t *action, lua_State *l)
+{
+    int (*func)(const action_t *a, lua_State *l);
+    func = action->func ?: default_function;
+    return func(action, l);
+}
