@@ -21,16 +21,15 @@
 static void export_as_png_slices(const char *path)
 {
     float box[4][4];
-    mesh_t *mesh;
+    const mesh_t *mesh;
     int x, y, z, w, h, d, pos[3], start_pos[3];
     uint8_t c[4];
     uint8_t *img;
     mesh_iterator_t iter = {0};
 
-    path = path ?: noc_file_dialog_open(NOC_FILE_DIALOG_SAVE,
-                   "png\0*.png\0", NULL, "untitled.png");
+    path = path ?: sys_get_save_path("png\0*.png\0", "untitled.png");
     if (!path) return;
-    mesh = goxel.layers_mesh;
+    mesh = goxel_get_layers_mesh();
     mat4_copy(goxel.image->box, box);
     if (box_is_null(box)) mesh_get_box(mesh, true, box);
     w = box[0][0] * 2;
@@ -53,8 +52,8 @@ static void export_as_png_slices(const char *path)
         img[(y * w * d + z * w + x) * 4 + 3] = c[3];
     }
     img_write(img, w * d, h, 4, path);
-
     free(img);
+    sys_on_saved(path);
 }
 
 ACTION_REGISTER(export_as_png_slices,
