@@ -129,16 +129,18 @@ static void export(const mesh_t *mesh, const char *path, bool ply)
         fprintf(out, "property float x\n");
         fprintf(out, "property float y\n");
         fprintf(out, "property float z\n");
-        fprintf(out, "property uchar red\n");
-        fprintf(out, "property uchar green\n");
-        fprintf(out, "property uchar blue\n");
+        fprintf(out, "property float red\n");
+        fprintf(out, "property float green\n");
+        fprintf(out, "property float blue\n");
         fprintf(out, "element face %d\n", utarray_len(lines_f));
-        fprintf(out, "property list uchar int vertex_index\n");
+        fprintf(out, "property list uchar int vertex_indices\n");
         fprintf(out, "end_header\n");
         while( (line_ptr = (line_t*)utarray_next(lines_v, line_ptr))) {
-            fprintf(out, "%g %g %g %d %d %d\n",
+            fprintf(out, "%g %g %g %f %f %f\n",
                     line_ptr->v[0], line_ptr->v[1], line_ptr->v[2],
-                    line_ptr->c[0], line_ptr->c[1], line_ptr->c[2]);
+                    line_ptr->c[0] / 255.,
+                    line_ptr->c[1] / 255.,
+                    line_ptr->c[2] / 255.);
         }
         while( (line_ptr = (line_t*)utarray_next(lines_f, line_ptr))) {
             if (size == 4) {
@@ -199,12 +201,22 @@ void ply_export(const mesh_t *mesh, const char *path)
 
 static void export_as_obj(const char *path)
 {
-    path = path ?: noc_file_dialog_open(NOC_FILE_DIALOG_SAVE,
-                    "obj\0*.obj\0", NULL, "untitled.obj");
+    path = path ?: sys_get_save_path("obj\0*.obj\0", "untitled.obj");
     if (!path) return;
-    wavefront_export(goxel.layers_mesh, path);
+    wavefront_export(goxel_get_layers_mesh(), path);
+    sys_on_saved(path);
 }
 
+ACTION_REGISTER(mesh_export_as_obj,
+    .help = "Export the mesh as a wavefront obj file",
+    .cfunc = wavefront_export,
+    .csig = "vpp",
+    .file_format = {
+        .name = "obj",
+        .ext = "*.obj\0",
+    },
+)
+
 ACTION_REGISTER(export_as_obj,
     .help = "Export the image as a wavefront obj file",
     .cfunc = export_as_obj,
@@ -217,12 +229,21 @@ ACTION_REGISTER(export_as_obj,
 
 static void export_as_ply(const char *path)
 {
-    path = path ?: noc_file_dialog_open(NOC_FILE_DIALOG_SAVE,
-                    "ply\0*.ply\0", NULL, "untitled.ply");
+    path = path ?: sys_get_save_path("ply\0*.ply\0", "untitled.ply");
     if (!path) return;
-    ply_export(goxel.layers_mesh, path);
+    ply_export(goxel_get_layers_mesh(), path);
 }
 
+ACTION_REGISTER(mesh_export_as_ply,
+    .help = "Export the mesh as a ply file",
+    .cfunc = ply_export,
+    .csig = "vpp",
+    .file_format = {
+        .name = "ply",
+        .ext = "*.ply\0",
+    },
+)
+
 ACTION_REGISTER(export_as_ply,
     .help = "Save the image as a ply file",
     .cfunc = export_as_ply,
