@@ -19,7 +19,11 @@
 // Procedural rendering inspired by ContextFree.org.
 // Need to write a doc about it.
 
+#include "procedural.h"
 #include "goxel.h"
+
+#include "utils/color.h"
+
 #include <stdarg.h>
 
 // Amount of time (s) after which we force stop a rendering iteration.
@@ -129,6 +133,14 @@ struct proc_ctx {
     bool        last;   // Mark the end of a frame.
 };
 
+/*
+ * Function: str_equ
+ * Return whether two strings are equal.
+ */
+static inline bool str_equ(const char *s1, const char *s2) {
+    return strcmp(s1, s2) == 0;
+}
+
 // Move a value toward a target.  If v is positive, move the value toward
 // range of v.  If v is negative move the value toward 0 of -v.
 // For example:
@@ -649,6 +661,58 @@ int proc_iter(gox_proc_t *proc, mesh_t *mesh, const painter_t *painter)
     return 0;
 }
 
+static int l_proc_run(const action_t *action, lua_State *l)
+{
+    gox_proc_t *proc;
+    mesh_t *mesh;
+    painter_t painter = (painter_t) {
+        .shape = &shape_cube,
+        .mode = MODE_INTERSECT,
+        .color = {255, 255, 255, 255},
+    };
+    proc = luaG_checkpointer(l, 1, "proc");
+    mesh = luaG_checkpointer(l, 2, "mesh");
+    proc_start(proc, NULL);
+    while (proc->state == PROC_RUNNING) {
+        proc_iter(proc, mesh, &painter);
+    }
+    return 0;
+}
+
+static gox_proc_t *proc_new(const char *code)
+{
+    gox_proc_t *proc;
+    proc = calloc(1, sizeof(*proc));
+    if (proc_parse(code, proc)) {
+        free(proc);
+        return NULL;
+    }
+    return proc;
+}
+
+static void proc_delete(gox_proc_t *proc)
+{
+    proc_release(proc);
+    free(proc);
+}
+
+ACTION_REGISTER(proc_new,
+    .help = "Create a new procedural prog",
+    .cfunc = proc_new,
+    .csig = "pp",
+)
+
+ACTION_REGISTER(proc_delete,
+    .help = "Delete a procedural prog",
+    .cfunc = proc_delete,
+    .csig = "vp",
+)
+
+ACTION_REGISTER(proc_run,
+    .help = "Run a procedural prog",
+    .func = l_proc_run,
+)
+
 // The actual parser code come here, generated from procedural.leg
 
 #ifdef __GNUC__
