@@ -16,11 +16,16 @@
  * goxel.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include "goxel.h"
+#include "system.h"
+
+#include "noc_file_dialog.h"
+
 #include <sys/time.h>
 #include <sys/stat.h>
 #include <dirent.h>
 #include <errno.h>
+#include <stdio.h>
+#include <string.h>
 
 #ifndef PATH_MAX
 #define PATH_MAX 1024
@@ -29,7 +34,7 @@
 // The global system instance.
 sys_callbacks_t sys_callbacks = {};
 
-#if defined(__unix__) && !defined(__EMSCRIPTEN__)
+#if defined(__unix__) && !defined(__EMSCRIPTEN__) && !defined(ANDROID)
 #define NOC_FILE_DIALOG_GTK
 #define NOC_FILE_DIALOG_IMPLEMENTATION
 #include "noc_file_dialog.h"
@@ -38,16 +43,16 @@ sys_callbacks_t sys_callbacks = {};
 
 static const char *get_user_dir(void *user)
 {
-    static char *ret = NULL;
+    static char ret[PATH_MAX] = "";
     const char *home;
-    if (!ret) {
+    if (!*ret) {
         home = getenv("XDG_CONFIG_HOME");
         if (home) {
-            asprintf(&ret, "%s/goxel", home);
+            snprintf(ret, sizeof(ret), "%s/goxel", home);
         } else {
             home = getenv("HOME");
             if (!home) home = getpwuid(getuid())->pw_dir;
-            asprintf(&ret, "%s/.config/goxel", home);
+            snprintf(ret, sizeof(ret), "%s/.config/goxel", home);
         }
     }
     return ret;
@@ -89,6 +94,9 @@ static void init_unix(void)
 #define NOC_FILE_DIALOG_IMPLEMENTATION
 #include "noc_file_dialog.h"
 
+// Defined in utils.
+int utf_16_to_8(const wchar_t *in16, char *out8, size_t size8);
+
 // On mingw mkdir takes only one argument!
 #define mkdir(p, m) mkdir(p)
 
@@ -118,10 +126,33 @@ static void init_win(void)
 #endif
 
 
+#ifdef __APPLE__
+#include <TargetConditionals.h>
+#if TARGET_OS_MAC
+
+const char *sys_get_save_path(const char *filters, const char *default_name)
+{
+    return noc_file_dialog_open(NOC_FILE_DIALOG_SAVE, filters, NULL,
+                                default_name);
+}
+
+void sys_on_saved(const char *path)
+{
+    LOG_I("Saved %s", path);
+}
+
+#endif
+#endif
+
+
 void sys_log(const char *msg)
 {
-    printf("%s\n", msg);
-    fflush(stdout);
+    if (sys_callbacks.log) {
+        sys_callbacks.log(sys_callbacks.user, msg);
+    } else {
+        printf("%s\n", msg);
+        fflush(stdout);
+    }
 }
 
 // List all the files in a directory.
@@ -168,7 +199,7 @@ int sys_make_dir(const char *path)
     return 0;
 }
 
-GLuint sys_get_screen_framebuffer(void)
+int sys_get_screen_framebuffer(void)
 {
     return 0;
 }
@@ -177,7 +208,7 @@ void sys_set_window_title(const char *title)
 {
     static char buf[1024] = {};
     if (strcmp(buf, title) == 0) return;
-    strncpy(buf, title, sizeof(buf));
+    snprintf(buf, sizeof(buf), "%s", title);
     if (sys_callbacks.set_window_title)
         sys_callbacks.set_window_title(sys_callbacks.user, title);
 }
@@ -208,3 +239,59 @@ void sys_set_clipboard_text(void *user, const char *text)
     if (sys_callbacks.set_clipboard_text)
         sys_callbacks.set_clipboard_text(sys_callbacks.user, text);
 }
+
+/*
+ * Function: sys_show_keyboard
+ * Show a virtual keyboard if needed.
+ */
+void sys_show_keyboard(bool has_text)
+{
+    if (!sys_callbacks.show_keyboard) return;
+    sys_callbacks.show_keyboard(sys_callbacks.user, has_text);
+}
+
+/*
+ * Function: sys_save_to_photos
+ * Save a png file to the system photo album.
+ */
+void sys_save_to_photos(const uint8_t *data, int size,
+                        void (*on_finished)(int r))
+{
+    FILE *file;
+    const char *path;
+    size_t r;
+
+    if (sys_callbacks.save_to_photos)
+        return sys_callbacks.save_to_photos(sys_callbacks.user, data, size,
+                                            on_finished);
+
+    // Default implementation.
+    path = noc_file_dialog_open(NOC_FILE_DIALOG_SAVE,
+                   "png\0*.png\0", NULL, "untitled.png");
+    if (!path) {
+        if (on_finished) on_finished(1);
+        return;
+    }
+    file = fopen(path, "wb");
+    if (!file) {
+        if (on_finished) on_finished(-1);
+        return;
+    }
+    r = fwrite(data, size, 1, file);
+    fclose(file);
+    if (on_finished) on_finished(r == size ? 0 : -1);
+}
+
+
+#ifdef NOC_FILE_DIALOG_IMPLEMENTATION
+const char *sys_get_save_path(const char *filters, const char *default_name)
+{
+    return noc_file_dialog_open(NOC_FILE_DIALOG_SAVE, filters, NULL,
+                                default_name);
+}
+
+void sys_on_saved(const char *path)
+{
+    LOG_I("Saved %s", path);
+}
+#endif
