@@ -22,7 +22,8 @@ typedef struct {
     tool_t tool;
 } tool_plane_t;
 
-static int iter(tool_t *tool, const float viewport[4])
+static int iter(tool_t *tool, const painter_t *painter,
+                const float viewport[4])
 {
     cursor_t *curs = &goxel.cursor;
     curs->snap_mask = SNAP_MESH;
@@ -35,6 +36,8 @@ static int iter(tool_t *tool, const float viewport[4])
         curs->pos[1] = round(curs->pos[1]);
         curs->pos[2] = round(curs->pos[2]);
         plane_from_normal(goxel.plane, curs->pos, curs->normal);
+        mat4_itranslate(goxel.plane, 0, 0, -1);
+        goxel.snap_mask |= SNAP_PLANE;
     }
     return 0;
 }
@@ -64,6 +67,7 @@ static int gui(tool_t *tool)
 }
 
 TOOL_REGISTER(TOOL_SET_PLANE, plane, tool_plane_t,
+              .name = "plane",
               .iter_fn = iter,
               .gui_fn = gui,
               .default_shortcut = "P"
